/****************************************************************
Golf program with simple planner.

Assumes walls are rectangles alingned with the x and y axis
****************************************************************/

#include <math.h>
#include <string.h>
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "texturepath.h"

/****************************************************************/

#ifdef _MSC_VER
#pragma warning(disable:4244 4305)  // for VC++, no precision loss complaints
#endif

/****************************************************************/

// You can choose the window size
#define WINDOW_WIDTH 700
#define WINDOW_HEIGHT 500

// Useful constants
#define TRUE 1
#define FALSE 0

/* Unfortunately, I had to decrease the simulation time
   step to keep the simulation stable */
#define TIME_STEP 0.0001

// object types
#define BALL 0
#define FLOOR 1
#define WALL 2
#define PLANE 3
#define ROBOT 4
#define FOOT 5
#define UNKNOWN -1

// golf sizes and masses.
#define BALL_MASS 0.04593
#define BALL_DIAMETER 0.0427
#define HOLE_RADIUS 0.04
#define FLOOR_HEIGHT (3*BALL_DIAMETER)
#define WALL_HEIGHT FLOOR_HEIGHT

#define SHAFT_HEIGHT 0.25
#define SHAFT_WIDTH 0.02
#define CLUB_HEAD_WIDTH 0.1
#define CLUB_HEAD_HEIGHT BALL_DIAMETER
#define CLUB_HEAD_THICKNESS 0.01

// putter head mass up to 0.5, MOI 13-46 oz-in^2
// http://www.golfworks.com/article.asp?ai=169
// Putters usually have a loft of 5° but cannot be more than 10°
// http://en.wikipedia.org/wiki/Putter

// various constants:

#define CLUB_POS_DESIRED (-0.2)

// velocity thresholds to implementing balls stopping quickly
#define APPLY_SUPER_DAMPING 1e-3
#define STOP_BALL 1e-4
#define IS_BALL_STOPPED 1e-5

// moving the ball if too close to walls parameters
#define NUDGE 0.001
#define MOVE_SIZE 0.1

#define FUDGE1 0.0001 // ball is actually launched in air
#define FUDGE2 0.0001 // space between floor tiles

// used to allocate space for floor tiles and walls
#define MAX_N_GEOMS 100

// used to allocate space for robot info.
#define MAX_N_ROBOT_PARTS 100

// robot sizes and masses
#define BODY_THICKNESS 0.1
#define BODY_WIDTH 0.3
#define HEAD_WIDTH 0.1
#define ARM_THICKNESS 0.05
#define ARM_WIDTH 0.1

#define FOOT_LENGTH 0.254
// #define FOOT_LENGTH 1.0 // Bigger foot is useful for debugging
#define FOOT_MASS 1.6
#define FOOT_THICKNESS 0.1
#define ANKLE_LOCATION 0.3 // fraction from rear of foot

#define SHANK_HEIGHT 0.412
#define SHANK_MASS 5.38

#define THIGH_HEIGHT 0.406
#define THIGH_MASS 13.04

#define TORSO_HEIGHT 0.762
#define TORSO_MASS 34.0

// STATE MACHINE
// enumerate states
#define PLAN 0
#define MOVING_CLUB_BACK 1
#define MOVING_INTO_SWING_POSITION 2
#define SWINGING 3
#define BALL_MOVING 4
#define STUCK 5

/****************************************************************/

// select correct drawing functions (we are using double version)
#ifdef dDOUBLE
#define dsDrawBox dsDrawBoxD
#define dsDrawSphere dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCapsule dsDrawCapsuleD
#define dsDrawConvex dsDrawConvexD
#define dsDrawTriangle dsDrawTriangleD
#endif

/****************************************************************/
// Typedefs

typedef struct rect
{
  double min_x;
  double min_y;
  double max_x;
  double max_y;
}
  RECT;

typedef struct hole
{
  RECT floor_size;

  dReal hole_x;
  dReal hole_y;

  dReal tee_x;
  dReal tee_y;

  // these are static ODE objects: just geometries
  dGeomID floor_tiles[MAX_N_GEOMS]; // ODE geoms
  int n_floor_tiles;
  RECT wall_rectangles[MAX_N_GEOMS]; // my representation of wall locations
  dGeomID the_walls[MAX_N_GEOMS]; // ODE geoms
  int n_walls;
}
  HOLE;

/****************************************************************/
// Declarations

static void nearCallback( void *data, dGeomID g1, dGeomID g2 );
void state_machine( HOLE *h );

/****************************************************************/
// Globals

// ODE dynamics, collision, joints, and other objects
dsFunctions fn;
static dWorldID world;
static dSpaceID space;
static dJointGroupID contact_joint_group;
static dJointGroupID robot_joint_group;
static dJointID club_joint;

static dReal time = 0; // what time is it

static int move_flag = 0;  // has ball just been moved?

static dGeomID plane_geom; // default floor for simulation

// the ODE ball
static dBodyID ball_body;
static dGeomID ball_geom;

// the robot
static dBodyID robot_bodies[MAX_N_ROBOT_PARTS];
static dGeomID robot_geoms[MAX_N_ROBOT_PARTS];
int n_robot_parts = 0;
static dJointID robot_joints[MAX_N_ROBOT_PARTS];
int n_robot_joints = 0;
static dReal joint_positions[MAX_N_ROBOT_PARTS];
static dReal joint_velocities[MAX_N_ROBOT_PARTS];
static dReal joint_torques[MAX_N_ROBOT_PARTS];
static dReal joint_positions_desired[MAX_N_ROBOT_PARTS];
static dReal k_pos[MAX_N_ROBOT_PARTS];
static dReal k_vel[MAX_N_ROBOT_PARTS];

// the current hole
HOLE current_hole;

// the current state in the state machine
int state = STUCK;

// the current desired position for the body
dReal body_goal_pos[3] = { 0.0, 0.0, 0.0 };

// the current feedforward torque
dReal club_joint_ff = 0;

// robot joints
int j_ankle = -1;
int j_knee = -1;
int j_hip = -1;
int j_neck = -1;
int j_shoulder = -1;
int j_elbow = -1;
int j_wrist = -1;

/****************************************************************/
/****************************************************************/
// Drawing, Simulation Loop, and other Callbacks
/****************************************************************/
// Draw the objects

void draw_stuff()
{
  int i;
  const dReal *pos;
  const dReal *R;
  dReal sides[3];

  // redraw sphere at new location
  pos = dGeomGetPosition( ball_geom );
  R = dGeomGetRotation( ball_geom );
  dsSetColor (1,0.5,0);
  dsDrawSphere( pos, R, dGeomSphereGetRadius( ball_geom ) );

  // redraw floor
  for ( i = 0; i < current_hole.n_floor_tiles; i++ )
    {
      pos = dGeomGetPosition( current_hole.floor_tiles[i] );
      R = dGeomGetRotation( current_hole.floor_tiles[i] );
      dGeomBoxGetLengths( current_hole.floor_tiles[i], sides );
      dsSetColor( 0.0, 0.75, 0.0 );
      dsDrawBox( pos, R, sides );
    }

  // redraw walls
  for ( i = 0; i < current_hole.n_walls; i++ )
    {
      pos = dGeomGetPosition( current_hole.the_walls[i] );
      R = dGeomGetRotation( current_hole.the_walls[i] );
      dGeomBoxGetLengths( current_hole.the_walls[i], sides );
      dsSetColor( 0.75, 0.5, 0.25 );
      dsDrawBox( pos, R, sides );
    }

  // redraw robot
  for ( i = 0; i < n_robot_parts; i++ )
    {
      pos = dGeomGetPosition( robot_geoms[i] );
      // printf( "pos: %g %g %g\n", pos[0], pos[1], pos[2] );
      R = dGeomGetRotation( robot_geoms[i] );
      dGeomBoxGetLengths( robot_geoms[i], sides );
      if ( (i % 2) == 0 )
	dsSetColor( 0.25, 0.25, 0.25 );
      else
	dsSetColor( 0.75, 0.75, 0.75 );
      dsDrawBox( pos, R, sides );
    }
}

/****************************************************************/
/*
  This is the main simulation loop that calls the collision detection
  function, steps the simulation, resets the temporary contact joint
  group, and redraws the objects at their new position.
*/

// simulation loop
static void simLoop (int pause)
{
  int i;

  state_machine( &current_hole );

  // find collisions and add contact joints
  dSpaceCollide( space, 0, &nearCallback );

  // step the simulation
  dWorldStep( world, TIME_STEP );  

  // increment time
  time += TIME_STEP;

  // remove all contact joints
  dJointGroupEmpty( contact_joint_group );

  draw_stuff();
}

/****************************************************************/
// used to tweak the ball physics to make it stop more quickly

void ball_move_callback( dBodyID body )
{
  int i;
  const dReal *vel;
  dReal vel_magnitude2 = 0;
  dReal damping = 0.99;

  vel = dBodyGetLinearVel( ball_body );
  for ( i = 0; i < 3; i++ )
    vel_magnitude2 += vel[i]*vel[i];

  // if ball is moving too slowly stop it.
  if ( vel_magnitude2 < STOP_BALL )
    {
      dBodySetLinearVel( ball_body, 0.0, 0.0, 0.0 );
    }
  else if ( vel_magnitude2 < APPLY_SUPER_DAMPING )
    {
      dBodySetLinearVel( ball_body,
			 damping*vel[0],
			 damping*vel[1],
			 damping*vel[2] );
    }
}

/****************************************************************/
// Used to figure out what kind of contact is going on.

void identify_geom( dGeomID g, int *type, const char **str, HOLE *h )
{
  int i;

  if ( g == plane_geom )
    {
      *type = PLANE;
      *str = "plane";
      return;
    }
  if ( g == ball_geom )
    {
      *type = BALL;
      *str = "ball";
      return;
    }
  for ( i = 0; i < h->n_floor_tiles; i++ )
    if ( g == h->floor_tiles[i] )
      {
	*type = FLOOR;
	*str = "floor";
	return;
      }
  for ( i = 0; i < h->n_walls; i++ )
    if ( g == h->the_walls[i] )
      {
	*type = WALL;
	*str = "wall";
	return;
      }
  for ( i = 0; i < n_robot_parts; i++ )
    if ( g == robot_geoms[i] )
      {
	*type = ROBOT;
	*str = "robot";
	return;
      }
  *type = UNKNOWN;
  *str = "unknown";
}

/****************************************************************/

static int ball_collision( dContact *c, int g_type, const char *g_string )
{

  if ( g_type == FLOOR )
    {
      c->surface.mode = dContactBounce | dContactSoftCFM;
      // friction parameter
      c->surface.mu = 0.5;
      // bounce is the amount of "bouncyness".
      c->surface.bounce = 0.1;
      // bounce_vel is the minimum incoming velocity to cause a bounce
      c->surface.bounce_vel = 0.01;
      // constraint force mixing parameter
      c->surface.soft_cfm = 0.01;  // 0.001 in bounce
      return TRUE;
    }

  if ( g_type == WALL )
    {
      c->surface.mode = dContactBounce | dContactSoftCFM;
      // friction parameter
      c->surface.mu = 1.0;
      // bounce is the amount of "bouncyness".
      c->surface.bounce = 0.9;
      // bounce_vel is the minimum incoming velocity to cause a bounce
      c->surface.bounce_vel = 0.01;
      // constraint force mixing parameter
      c->surface.soft_cfm = 0.01;  // 0.001 in bounce
      return TRUE;
    }

  if ( g_type == ROBOT )
    {
      c->surface.mode = dContactBounce | dContactSoftCFM;
      // friction parameter
      c->surface.mu = 1.0;
      // bounce is the amount of "bouncyness".
      c->surface.bounce = 0.9;
      // bounce_vel is the minimum incoming velocity to cause a bounce
      c->surface.bounce_vel = 0.01;
      // constraint force mixing parameter
      c->surface.soft_cfm = 0.01;  // 0.001 in bounce
      return TRUE;
    }

  if ( g_type == PLANE )
    {
      c->surface.mode = dContactBounce | dContactSoftCFM;
      // friction parameter
      c->surface.mu = dInfinity;
      // bounce is the amount of "bouncyness".
      c->surface.bounce = 0.5;
      // bounce_vel is the minimum incoming velocity to cause a bounce
      c->surface.bounce_vel = 0.1;
      // constraint force mixing parameter
      c->surface.soft_cfm = 0.01;  // 0.001 in bounce
      return TRUE;
    }

  return FALSE; // no floor or wall involved, give up
}

/****************************************************************/

static int robot_collision( dContact *c, int g_type, const char *g_string )
{

  if ( g_type == FLOOR )
    {
      // c->surface.mode = 0;
      c->surface.mode = dContactSoftERP | dContactSoftCFM;
      // friction parameter
      c->surface.mu = dInfinity;
      // error reduction parameter
      c->surface.soft_erp = 0.3; // 0.1 to 0.8, 0.2 is default
      // constraint force mixing parameter
      c->surface.soft_cfm = 0.001;  // 0.001 in bounce
      return TRUE;
    }

  if ( g_type == PLANE )
    {
      // c->surface.mode = 0;
      c->surface.mode = dContactSoftERP | dContactSoftCFM;
      // friction parameter
      c->surface.mu = dInfinity;
      // error reduction parameter
      c->surface.soft_erp = 0.3; // 0.1 to 0.8, 0.2 is default
      // constraint force mixing parameter
      c->surface.soft_cfm = 0.001;  // 0.001 in bounce
      return TRUE;
    }

  return FALSE; // no floor involved, give up
}

/****************************************************************/
/*
When the collision system detects that two objects are potentially colliding,
it calls this routine which determines the points of contact and creates
temporary joints. The surface parameters of the joint (friction,
bounce velocity, CFM, etc) are also set here.
*/

/*
What makes this complicated is that we want to detect Ball collisions
with {FLOOR, WALL, ROBOT} and robot collisions with the floor.
We ignore other collisions.
*/

static void nearCallback( void *data, dGeomID g1, dGeomID g2 )
{
  dBodyID b1 = dGeomGetBody( g1 );
  dBodyID b2 = dGeomGetBody( g2 );
  dContact contact;
  int g1_type, g2_type, g_type;
  const char *g1_string, *g2_string, *g_string;

  // Need to identify geom:
  identify_geom( g1, &g1_type, &g1_string, &current_hole );
  identify_geom( g2, &g2_type, &g2_string, &current_hole );
  // printf( "%s %s\n", g1_string, g2_string );
  // give up if unknown object involved
  if ( g1_type == UNKNOWN )
    return;
  if ( g2_type == UNKNOWN )
    return;
  if ( g1_type == BALL )
    {
      if ( !ball_collision( &contact, g2_type, g2_string ) )
	return;
    }
  else if ( g2_type == BALL )
    {
      if ( !ball_collision( &contact, g1_type, g1_string ) )
	return;
    }
  else if ( g1_type == ROBOT )
    {
      if ( !robot_collision( &contact, g2_type, g2_string ) )
	return;
    }
  else if ( g2_type == ROBOT )
    {
      if ( !robot_collision( &contact, g1_type, g1_string ) )
	return;
    }
  else
    return; // no ball or robot involved, give up.

  if ( int numc = dCollide( g1, g2, 1, &contact.geom, sizeof(dContact) ) )
    {
      dJointID c = dJointCreateContact( world, contact_joint_group, &contact );
      dJointAttach( c, b1, b2 );
    }
}

/****************************************************************/
// called when a key pressed

static void command (int cmd)
{
  float vx, vy;

  // this is a leftover from past demo programs
  /*
    if ( cmd == 's' )
    {
    printf( "s pressed.\n" );
    const dReal *pos = dBodyGetPosition( ball_body );
    printf( "ball at %g %g %g; hole at %g %g\n",
    pos[0], pos[1], pos[2],
    current_hole.hole_x, current_hole.hole_y );
    printf( "enter ball velocity as two numbers: -0.05 -0.4\n" );
    scanf( "%g%g", &vx, &vy );
    printf( "you typed: %g %g\n", vx, vy );
    // lift ball a little bit to get it to move
    dBodySetPosition( ball_body, pos[0], pos[1], pos[2] + FUDGE1 );
    dBodySetLinearVel( ball_body, (dReal) vx, (dReal) vy, 0.0 );
    }
  */
}

/****************************************************************/
/****************************************************************/
// SET THINGS UP
/****************************************************************/
// used to figure out how big the hole area is

void update_min_max( HOLE *h, dReal x, dReal y )
{

  if ( x < h->floor_size.min_x )
    h->floor_size.min_x = x;
  if ( x > h->floor_size.max_x )
    h->floor_size.max_x = x;

  if ( y < h->floor_size.min_y )
    h->floor_size.min_y = y;
  if ( y > h->floor_size.max_y )
    h->floor_size.max_y = y;
}

/****************************************************************/

int read_hole_file( char *filename, HOLE *h )
{
  int i;
  FILE *stream;
  char buffer[1000];

  stream = fopen( filename, "r" );
  if ( stream == NULL )
    {
      fprintf( stderr, "Can't open hole file %s", filename );
      exit( -1 );
    }

  h->floor_size.min_x = 1e10;
  h->floor_size.min_y = 1e10;
  h->floor_size.max_x = -1e10;
  h->floor_size.max_y = -1e10;
  h->n_floor_tiles = 0;
  h->n_walls = 0;

  for( ; ; )
    {
      // read keyword
      if ( fscanf( stream, "%s", buffer ) < 1 )
	break; // if we didn't read anything we are done

      // handle a hole x y
      if ( strcmp( buffer, "hole" ) == 0 )
	{
	  if ( fscanf( stream, "%lg%lg", &(h->hole_x), &(h->hole_y) ) < 2 )
	    {
	      fprintf( stderr, "bad hole in hole file %s\n", filename );
	      exit( -1 );
	    }
	  update_min_max( h, h->hole_x, h->hole_y ); 
	  printf( "hole: %g %g\n", h->hole_x, h->hole_y );
	  continue;
	}
      
      // handle a tee x y
      if ( strcmp( buffer, "tee" ) == 0 )
	{
	  if ( fscanf( stream, "%lg%lg", &(h->tee_x), &(h->tee_y) ) < 2 )
	    {
	      fprintf( stderr, "bad tee in hole file %s\n", filename );
	      exit( -1 );
	    }
	  update_min_max( h, h->tee_x, h->tee_y ); 
	  printf( "tee: %g %g\n", h->tee_x, h->tee_y );
	  continue;
	}

      // handle a wall min_x min_y max_x max_y
      if ( strcmp( buffer, "wall" ) == 0 )
	{
	  i = h->n_walls;
	  if ( fscanf( stream, "%lg%lg%lg%lg",
		       &(h->wall_rectangles[i].min_x),
		       &(h->wall_rectangles[i].min_y),
		       &(h->wall_rectangles[i].max_x),
		       &(h->wall_rectangles[i].max_y) ) < 4 )
	    {
	      fprintf( stderr, "bad wall in hole file %s\n", filename );
	      exit( -1 );
	    }
	  update_min_max( h,
			  h->wall_rectangles[i].min_x,
			  h->wall_rectangles[i].min_y );
	  update_min_max( h,
			  h->wall_rectangles[i].max_x,
			  h->wall_rectangles[i].max_y );
	  printf( "wall: %g %g %g %g\n",
		  h->wall_rectangles[i].min_x,
		  h->wall_rectangles[i].min_y,
		  h->wall_rectangles[i].max_x,
		  h->wall_rectangles[i].max_y );
	  h->n_walls++;
	  continue;
	}
      
      fprintf( stderr, "bad keyword %s in hole file %s\n", buffer, filename );
      exit( -1 );
    }

  fclose( stream );
}

/****************************************************************/

void create_bodies_and_geoms( char *filename, HOLE *h )
{
  int i, j;
  dMass m;

  read_hole_file( filename, h );

  // create ball object
  ball_body = dBodyCreate( world );
  ball_geom = dCreateSphere( space, BALL_DIAMETER/2 );
  dMassSetSphereTotal( &m, BALL_MASS, BALL_DIAMETER/2 );
  dBodySetMass( ball_body, &m );
  dGeomSetBody( ball_geom, ball_body );
  // set initial position and velocity
  dBodySetPosition( ball_body, h->tee_x, h->tee_y,
		    FLOOR_HEIGHT + BALL_DIAMETER/2 + FUDGE1 );
  dBodySetLinearDamping( ball_body, 1e-4 ); 
  dBodySetLinearDampingThreshold( ball_body, 1e-7 );
  dBodySetMovedCallback( ball_body, ball_move_callback );

  // create robot parts
  // FOOT
  i = 0;
  robot_bodies[i] = dBodyCreate( world );
  robot_geoms[i] = dCreateBox( space,
			       FOOT_LENGTH,
			       BODY_WIDTH,
			       FOOT_THICKNESS );
  dMassSetBoxTotal( &m, FOOT_MASS, FOOT_LENGTH, BODY_WIDTH, FOOT_THICKNESS );
  dBodySetMass( robot_bodies[i], &m );
  dGeomSetBody( robot_geoms[i], robot_bodies[i] );
  dBodySetPosition( robot_bodies[i],
		    -ANKLE_LOCATION*FOOT_LENGTH + FOOT_LENGTH/2,
		    0.0,
		    FOOT_THICKNESS/2.0 + FLOOR_HEIGHT );

  // SHANK
  i++;
  robot_bodies[i] = dBodyCreate( world );
  robot_geoms[i] = dCreateBox( space,
			       BODY_THICKNESS,
			       BODY_WIDTH,
			       SHANK_HEIGHT );
  dMassSetBoxTotal( &m, SHANK_MASS, BODY_THICKNESS, BODY_WIDTH, SHANK_HEIGHT );
  dBodySetMass( robot_bodies[i], &m );
  dGeomSetBody( robot_geoms[i], robot_bodies[i] );
  dBodySetPosition( robot_bodies[i], 0.0, 0.0,
		    SHANK_HEIGHT/2 + FOOT_THICKNESS + FLOOR_HEIGHT );

  j_ankle = j = n_robot_joints++;
  robot_joints[j] = dJointCreateHinge( world, robot_joint_group );
  dJointAttach( robot_joints[j], robot_bodies[i-1], robot_bodies[i] );
  dJointSetHingeAnchor( robot_joints[j], 0.0, 0.0,
			FOOT_THICKNESS + FLOOR_HEIGHT );
  dJointSetHingeAxis( robot_joints[j], 0.0, 1.0, 0.0 );  

  // THIGH
  i++;
  robot_bodies[i] = dBodyCreate( world );
  robot_geoms[i] = dCreateBox( space,
			       BODY_THICKNESS,
			       BODY_WIDTH,
			       THIGH_HEIGHT );
  dMassSetBoxTotal( &m, THIGH_MASS, BODY_THICKNESS, BODY_WIDTH, THIGH_HEIGHT );
  dBodySetMass( robot_bodies[i], &m );
  dGeomSetBody( robot_geoms[i], robot_bodies[i] );
  dBodySetPosition( robot_bodies[i], 0.0, 0.0,
		    THIGH_HEIGHT/2 + SHANK_HEIGHT + FOOT_THICKNESS
		    + FLOOR_HEIGHT );

  j_knee = j = n_robot_joints++;
  robot_joints[j] = dJointCreateHinge( world, robot_joint_group );
  dJointAttach( robot_joints[j], robot_bodies[i-1], robot_bodies[i] );
  dJointSetHingeAnchor( robot_joints[j], 0.0, 0.0,
			SHANK_HEIGHT + FOOT_THICKNESS + FLOOR_HEIGHT );
  dJointSetHingeAxis( robot_joints[j], 0.0, 1.0, 0.0 );  

  // TORSO
  i++;
  robot_bodies[i] = dBodyCreate( world );
  robot_geoms[i] = dCreateBox( space,
			       BODY_THICKNESS,
			       BODY_WIDTH,
			       TORSO_HEIGHT );
  dMassSetBoxTotal( &m, TORSO_MASS, BODY_THICKNESS, BODY_WIDTH, TORSO_HEIGHT );
  dBodySetMass( robot_bodies[i], &m );
  dGeomSetBody( robot_geoms[i], robot_bodies[i] );
  dBodySetPosition( robot_bodies[i], 0.0, 0.0,
		    TORSO_HEIGHT/2 + THIGH_HEIGHT + SHANK_HEIGHT
		    + FOOT_THICKNESS + FLOOR_HEIGHT );

  j_hip = j = n_robot_joints++;
  robot_joints[j] = dJointCreateHinge( world, robot_joint_group );
  dJointAttach( robot_joints[j], robot_bodies[i-1], robot_bodies[i] );
  dJointSetHingeAnchor( robot_joints[j], 0.0, 0.0,
			THIGH_HEIGHT + SHANK_HEIGHT + FOOT_THICKNESS
			+ FLOOR_HEIGHT );
  dJointSetHingeAxis( robot_joints[j], 0.0, 1.0, 0.0 );  

#ifdef COMMENT
  // commented out until the humanoid starts to play golf.
  // CLUB SHAFT
  i = 1;
  robot_bodies[i] = dBodyCreate( world );
  robot_geoms[i] = dCreateBox( space,
			       CLUB_HEAD_THICKNESS,
			       CLUB_HEAD_WIDTH,
			       SHAFT_HEIGHT );
  dMassSetBox( &m, 7800.0, SHAFT_WIDTH, CLUB_HEAD_WIDTH, SHAFT_HEIGHT );
  dBodySetMass( robot_bodies[i], &m );
  printf( "Club mass: %g\n", m.mass );
  dGeomSetBody( robot_geoms[i], robot_bodies[i] );
  dBodySetPosition( robot_bodies[i], h->tee_x, h->tee_y - BODY_WIDTH/2,
		    BODY_HEIGHT_UP - SHAFT_HEIGHT/2 /* - CLUB_HEAD_HEIGHT */ );

  club_joint = dJointCreateHinge( world, robot_joint_group );
  dJointAttach( club_joint, robot_bodies[0], robot_bodies[1] );
  dJointSetHingeAnchor( club_joint, h->tee_x, h->tee_y - BODY_WIDTH/2,
			BODY_HEIGHT_UP );
  dJointSetHingeAxis( club_joint, 0.0, 1.0, 0.0 );  
#endif

  n_robot_parts = i + 1;

  /*
    FROM FAQ:

  How can an immovable body be created?

  In other words, how can you create a body that doesn't move, but
  that interacts with other bodies? The answer is to create a geom
  only, without the corresponding rigid body object. The geom is
  associated with a rigid body ID of zero. Then in the contact
  callback when you detect a collision between two geoms with a
  nonzero body ID and a zero body ID, you can simply pass those two
  IDs to the dJointAttach function as normal. This will create a
  contact between the rigid body and the static environment.
  */

  // create floor objects

  printf( "%g %g\n",
	  h->hole_x - h->floor_size.min_x - HOLE_RADIUS - FUDGE2,
	  h->floor_size.max_y - h->floor_size.min_y );
  h->floor_tiles[0] = dCreateBox( space,
			h->hole_x - h->floor_size.min_x - HOLE_RADIUS - FUDGE2,
				h->floor_size.max_y - h->floor_size.min_y,
				FLOOR_HEIGHT );
  dGeomSetPosition( h->floor_tiles[0],
		    h->hole_x - HOLE_RADIUS
		    - (h->hole_x - h->floor_size.min_x - HOLE_RADIUS)/2.0,
		    (h->floor_size.max_y + h->floor_size.min_y)/2.0,
		    FLOOR_HEIGHT/2.0 + FUDGE2 );

  printf( "%g %g\n",
	  h->floor_size.max_x - h->hole_x - HOLE_RADIUS - FUDGE2,
	  h->floor_size.max_y - h->floor_size.min_y );
  h->floor_tiles[1] = dCreateBox( space,
			h->floor_size.max_x - h->hole_x - HOLE_RADIUS - FUDGE2,
				h->floor_size.max_y - h->floor_size.min_y,
				FLOOR_HEIGHT );
  dGeomSetPosition( h->floor_tiles[1],
		    h->hole_x + HOLE_RADIUS
		    + (h->floor_size.max_x - h->hole_x - HOLE_RADIUS)/2.0,
		    (h->floor_size.max_y + h->floor_size.min_y)/2.0,
		    FLOOR_HEIGHT/2.0 + FUDGE2 );

  printf( "%g %g\n",
	  2*HOLE_RADIUS,
	  h->hole_y - h->floor_size.min_y - HOLE_RADIUS );
  h->floor_tiles[2] = dCreateBox( space,
				2*HOLE_RADIUS,
				h->hole_y - h->floor_size.min_y - HOLE_RADIUS,
				FLOOR_HEIGHT );
  dGeomSetPosition( h->floor_tiles[2],
		    h->hole_x,
		    h->hole_y - HOLE_RADIUS
		    - (h->hole_y - h->floor_size.min_y - HOLE_RADIUS)/2,
		    FLOOR_HEIGHT/2 + FUDGE2 );

  printf( "%g %g\n",
	  2*HOLE_RADIUS,
	  h->floor_size.max_y - h->hole_y - HOLE_RADIUS );
  h->floor_tiles[3] = dCreateBox( space,
				2*HOLE_RADIUS,
				h->floor_size.max_y - h->hole_y - HOLE_RADIUS,
				FLOOR_HEIGHT );
  dGeomSetPosition( h->floor_tiles[3],
		    h->hole_x,
		    h->hole_y + HOLE_RADIUS
		    + (h->floor_size.max_y - h->hole_y - HOLE_RADIUS)/2,
		    FLOOR_HEIGHT/2 + FUDGE2 );

  h->n_floor_tiles = 4;

  // create walls

  for ( i = 0; i < h->n_walls; i++ )
    {
      printf( "%g %g\n",
	      h->wall_rectangles[i].max_x - h->wall_rectangles[i].min_x,
	      h->wall_rectangles[i].max_y - h->wall_rectangles[i].min_y );
      h->the_walls[i] =
	dCreateBox( space,
		    h->wall_rectangles[i].max_x - h->wall_rectangles[i].min_x,
		    h->wall_rectangles[i].max_y - h->wall_rectangles[i].min_y,
		    WALL_HEIGHT );
      dGeomSetPosition( h->the_walls[i],
		(h->wall_rectangles[i].max_x + h->wall_rectangles[i].min_x)/2,
		(h->wall_rectangles[i].max_y + h->wall_rectangles[i].min_y)/2,
		FLOOR_HEIGHT + WALL_HEIGHT/2 + 2*FUDGE2 );
    }

  // plane that catches balls that leave course
  plane_geom = dCreatePlane( space, 0, 0, 1, 0 );
}


/****************************************************************/

// This function is called at the start of the simulation
// It sets up the viewpoint, and anything else you might want.
static void start()
{
  float xyz[3] = { 1.0f, 2.3f, 1.0f };
  float hpr[3] = { -110.0f, -7.0f, 0.0f };

  dsSetViewpoint( xyz, hpr );
}

/****************************************************************/
/****************************************************************/
// Handle robot actions
/****************************************************************/
/****************************************************************/
// Move the ball if necessary
// move the ball away from walls, and nudge it a little bit.
// this will have problems with multiple close walls or walls at
// non-right angles

int move_ball( dReal *pos, HOLE *h )
{
  int i, j;
  const dReal *wall_pos;
  const dReal *wall_R;
  dReal sides[3];
  int move_count = 0;

  // if just moved don't move it again
  if ( move_flag )
    return 0;

  // printf( "pos1: %g %g\n", pos[0], pos[1] );

  // first randomly move the ball
  for ( i = 0; i < 2; i++ )
    {
      pos[i] += 2*NUDGE*dRandReal() - NUDGE;
    }

  // printf( "pos2: %g %g\n", pos[0], pos[1] );

  // now move it away from any wall
  for ( i = 0; i < h->n_walls; i++ )
    {
      wall_pos = dGeomGetPosition( h->the_walls[i] );
      wall_R = dGeomGetRotation( h->the_walls[i] );
      dGeomBoxGetLengths( h->the_walls[i], sides );
      // ignore rotation R for now.
      for ( j = 0; j < 2; j++ )
	{
	  // make the lengths half lengths, and add the offset
	  sides[j] = sides[j]/2 + MOVE_SIZE;
	}
	  // check if close enough
      if ( fabs( (double) wall_pos[0] - pos[0] ) < sides[0]
	   && fabs( (double) wall_pos[1] - pos[1] ) < sides[1] )
	{
	  if ( sides[0] < sides[1] )
	    {
	      j = 0;
	      /*
	      printf( "contact %d %g %g %g\n",
		      j, pos[j], h->wall_pos[j], sides[j] );
	      */
	    }
	  else
	    {
	      j = 1;
	      /*
	      printf( "contact %d %g %g %g\n",
		      j, pos[j], h->wall_pos[j], sides[j] );
	      */
	    }
	  move_count++;
	  if ( pos[j] > wall_pos[j] )
	    pos[j] = wall_pos[j] + MOVE_SIZE;
	  else
	    pos[j] = wall_pos[j] - MOVE_SIZE;
	  // printf( "Too close: %d %d %g %g\n", i, j, wall_pos[j], pos[j] );
	}
    }
  if ( move_count > 0 )
    printf( "%g: ball moved to %g %g\n", time, pos[0], pos[1] );

  // lift ball a little bit to get it to move
  dBodySetPosition( ball_body, pos[0], pos[1], pos[2] + FUDGE1 );

  if ( move_count > 0 )
    move_flag = 1;

  return move_count;
}

/****************************************************************/
/****************************************************************/
/****************************************************************/
/****************************************************************/
/****************************************************************/
/****************************************************************/
/****************************************************************/
/****************************************************************/

void plan( HOLE *h )
{
  int i;
  const dReal *pos;
  const dReal *vel;
  dReal vel_magnitude2 = 0;
  dReal pos_magnitude2 = 0;
  dReal dx, dy, vx, vy, d_magnitude;
  static dReal last_pos[3];
  dReal new_pos[3];
  static dReal wait_until = 5.0;
  dReal *R;
  dReal array[100];
  dReal sin_angle, cos_angle;

  // I don't understand C++
  R = (dReal *) array;

  pos = dBodyGetPosition( ball_body );

  if ( pos[2] < FLOOR_HEIGHT )
    {
      printf( "%g: ball in hole at %g %g %g; hole at %g %g\n", time,
	      pos[0], pos[1], pos[2], h->hole_x, h->hole_y );
      state = STUCK;
      return;
    }

  // return if we are waiting (allows viewer to see ball stop).
  /*
  if ( time < wait_until )
    return;
  */

  printf( "%g: ball at %g %g %g; hole at %g %g\n", time,
	  pos[0], pos[1], pos[2], h->hole_x, h->hole_y );

  for ( i = 0; i < 3; i++ )
    pos_magnitude2 += (pos[i] - last_pos[i])*(pos[i] - last_pos[i]);
  if ( pos_magnitude2 < 1e-7 && move_flag == 0 )
    {
      if ( state != STUCK )
	{
	  state = STUCK;
	  printf( "%g: I am stuck %g\n", time, pos_magnitude2 );
	  return;
	}
      else
	return;
    }

  // move the ball away from walls, and nudge it a little bit.
  for ( i = 0; i < 3; i++ )
    new_pos[i] = pos[i];

  if ( move_ball( new_pos, h ) > 0 )
    {
      for ( i = 0; i < 3; i++ )
	last_pos[i] = new_pos[i];
      wait_until = time + 5.0;
      return;
    }

  for ( i = 0; i < 3; i++ )
    last_pos[i] = new_pos[i];

  dx = h->hole_x - new_pos[0];
  dy = h->hole_y - new_pos[1];
  d_magnitude = sqrt(dx*dx + dy*dy); // limit max velocity to 0.33
  /*
  vx = dx/d_magnitude;
  vy = dy/d_magnitude;
  */

  // zzzzzzzzzzzzzzzzzzzzz

  printf( "%g: hit direction: %g %g\n", time, dx, dy );

  dRFromAxisAndAngle( R, 0.0, 0.0, 1.0, (dReal) atan2( dy, dx ) );
  // dBodySetRotation( robot_bodies[ROOT], R );

  sin_angle = dy/d_magnitude;
  cos_angle = dx/d_magnitude;

  body_goal_pos[0] = new_pos[0] - 0.5*BODY_WIDTH*sin_angle
    -0.5*(BALL_DIAMETER + CLUB_HEAD_THICKNESS)*cos_angle;
  body_goal_pos[1] = new_pos[1] + 0.5*BODY_WIDTH*cos_angle
    -0.5*(BALL_DIAMETER + CLUB_HEAD_THICKNESS)*sin_angle;

  printf( "ball: %g %g; body desired: %g %g; %g %g\n", new_pos[0], new_pos[1],
	  body_goal_pos[0], body_goal_pos[1] );

  printf( "angle: %g; sin: %g cos: %g\n",
	  atan2( dy, dx ), sin_angle, cos_angle );
  /*
  dBodySetPosition( robot_bodies[ROOT], new_pos[0], new_pos[1],
		    ROBOT_HEIGHT/2.0 );
  */

  /*
  dBodySetLinearVel( ball_body, (dReal) vx, (dReal) vy, 0.0 );
  */

  move_flag = 0;
  state = MOVING_CLUB_BACK; // reset state machine
}

/****************************************************************/

void move_club_back_waiting( HOLE *h )
{
  const dReal *body_pos;
  dReal pos;
  dReal vel;
  dReal k_pos = 10.0;
  dReal k_vel = 2.0;
  dReal k_int = 0.01;
  dReal pos_desired = CLUB_POS_DESIRED;
  dReal torque;

#ifdef COMMENT
  // commented out until the humanoid starts to play golf.
  // Move body
  body_pos = dBodyGetPosition( robot_bodies[ROOT] );
  if ( body_pos[2] < BODY_HEIGHT_UP )
    dBodySetLinearVel( robot_bodies[ROOT], 0.0, 0.0,
		       MAX_BODY_VERTICAL_SPEED );
  else
    dBodySetLinearVel( robot_bodies[ROOT], 0.0, 0.0, 0.0 );

  // Move club
  pos = dJointGetHingeAngle( club_joint );
  vel = dJointGetHingeAngleRate( club_joint );
  club_joint_ff -= k_int*( pos - pos_desired );

  torque = - k_pos*( pos - pos_desired ) - k_vel*vel + club_joint_ff;
  dJointAddHingeTorque( club_joint, torque );
#endif

  // no transition out, wait() does that.
}

/****************************************************************/

void move_club_back( HOLE *h )
{
  int i;
  const dReal *body_pos = NULL;
  dReal pos = 0.0;
  dReal vel = 0.0;
  dReal k_pos = 10.0;
  dReal k_vel = 2.0;
  dReal k_int = 0.1;
  dReal pos_desired = CLUB_POS_DESIRED;
  dReal torque;
  dReal body_vel[3];
  dReal vel_magnitude2 = 0.0;

#ifdef COMMENT
  // commented out until the humanoid starts to play golf.
  // Move body
  body_pos = dBodyGetPosition( robot_bodies[ROOT] );
  for ( i = 0; i < 2; i++ )
    {
      body_vel[i] = body_goal_pos[i] - body_pos[i];
      if ( body_vel[i] > MAX_BODY_SPEED )
	body_vel[i] = MAX_BODY_SPEED;
      else if ( body_vel[i] < -MAX_BODY_SPEED )
	body_vel[i] = -MAX_BODY_SPEED;
    }
  if ( body_pos[2] < BODY_HEIGHT_UP )
    body_vel[2] = MAX_BODY_VERTICAL_SPEED;
  else
    body_vel[2] = 0.0;

  dBodySetLinearVel( robot_bodies[ROOT],
		     body_vel[0], body_vel[1], body_vel[2] );

  for ( i = 0; i < 3; i++ )
    vel_magnitude2 += body_vel[i]*body_vel[i];

  // Move club
  pos = dJointGetHingeAngle( club_joint );
  vel = dJointGetHingeAngleRate( club_joint );
  club_joint_ff -= k_int*( pos - pos_desired );

  torque = - k_pos*( pos - pos_desired ) - k_vel*vel + club_joint_ff;
  dJointAddHingeTorque( club_joint, torque );

  /*
  printf( "%g %g; %g %g %g\n", body_pos[2], BODY_HEIGHT_UP,
	  pos, vel, torque );
  */

  if ( vel_magnitude2 < 0.00001
       && ( pos > pos_desired - 0.001 ) )
    {
      dBodySetLinearVel( robot_bodies[ROOT], 0.0, 0.0, 0.0 );
      printf( "Club back: %g %g %g %g; %g %g %g\n",
	      body_pos[0], body_pos[1], body_pos[2], BODY_HEIGHT_UP,
	      pos, vel, torque );
      state = MOVING_INTO_SWING_POSITION;
    }
#endif
}

/****************************************************************/

void move_into_swing_position( HOLE *h )
{
  int i;
  const dReal *body_pos = NULL;
  dReal pos = 0.0;
  dReal vel = 0.0;
  dReal k_pos = 10.0;
  dReal k_vel = 2.0;
  dReal k_int = 0.1;
  dReal pos_desired = CLUB_POS_DESIRED;
  dReal torque = 0.0;
  dReal body_vel[3];
  dReal vel_magnitude2 = 0.0;

#ifdef COMMENT
  // commented out until the humanoid starts to play golf.
  // Move body
  body_pos = dBodyGetPosition( robot_bodies[ROOT] );
  for ( i = 0; i < 2; i++ )
    {
      body_vel[i] = body_goal_pos[i] - body_pos[i];
      if ( body_vel[i] > MAX_BODY_SPEED )
	body_vel[i] = MAX_BODY_SPEED;
      else if ( body_vel[i] < -MAX_BODY_SPEED )
	body_vel[i] = -MAX_BODY_SPEED;
    }
  i = 2;
  body_vel[i] = BODY_HEIGHT_DOWN - body_pos[i];
  if ( body_vel[i] > MAX_BODY_VERTICAL_SPEED )
    body_vel[i] = MAX_BODY_VERTICAL_SPEED;
  else if ( body_vel[i] < -MAX_BODY_VERTICAL_SPEED )
    body_vel[i] = -MAX_BODY_VERTICAL_SPEED;

  dBodySetLinearVel( robot_bodies[ROOT],
		     body_vel[0], body_vel[1], body_vel[2] );

  for ( i = 0; i < 3; i++ )
    vel_magnitude2 += body_vel[i]*body_vel[i];

  // Move club
  pos = dJointGetHingeAngle( club_joint );
  vel = dJointGetHingeAngleRate( club_joint );
  club_joint_ff -= k_int*( pos - pos_desired );

  torque = - k_pos*( pos - pos_desired ) - k_vel*vel + club_joint_ff;
  dJointAddHingeTorque( club_joint, torque );

  // printf( "%g %g %g\n", pos, vel, torque );

  if ( vel_magnitude2 < 1e-10 )
    {
      dBodySetLinearVel( robot_bodies[ROOT], 0.0, 0.0, 0.0 );
      printf( "Club in swing pos: %g %g %g %g; %g %g %g\n",
	      body_pos[0], body_pos[1], body_pos[2], BODY_HEIGHT_DOWN,
	      pos, vel, torque );
      state = SWINGING;
    }
#endif
}

/****************************************************************/

void swing_club( HOLE *h )
{
  const dReal *body_pos = NULL;
  dReal pos = 0.0;
  dReal vel = 0.0;
  dReal k_pos = 1.0e-3;
  dReal k_vel = 1.0e-3;
  dReal k_int = 1.0e-6;
  dReal pos_desired = -CLUB_POS_DESIRED;
  dReal torque = 0;

#ifdef COMMENT
  // commented out until the humanoid starts to play golf.
  // Move club
  pos = dJointGetHingeAngle( club_joint );
  vel = dJointGetHingeAngleRate( club_joint );
  // club_joint_ff -= k_int*( pos - pos_desired );

  // torque = -club_joint_ff;
  // dJointAddHingeTorque( club_joint, torque );
  dJointAddHingeTorque( club_joint, 0.0 );

  // printf( "%g %g %g\n", pos, vel, torque );

  if ( fabs( vel ) < 1.0
       && pos > pos_desired/2 )
    {
      club_joint_ff = 0;
      printf( "Swing done: %g %g %g\n",
	      pos, vel, torque );
      state = BALL_MOVING;
    }
#endif
}

/****************************************************************/

void ball_moving( HOLE *h )
{
  int i;
  const dReal *vel;
  dReal vel_magnitude2 = 0;

  // Is ball moving?
  vel = dBodyGetLinearVel( ball_body );
  for ( i = 0; i < 3; i++ )
    vel_magnitude2 += vel[i]*vel[i];

  // if ball is moving return
  if ( vel_magnitude2 < IS_BALL_STOPPED )
    {
      printf( "Ball stopped: %g; %g %g %g\n", 
	      vel_magnitude2, vel[0], vel[1], vel[2] );
      state = PLAN;
      return;
    }
}

/****************************************************************/

void stuck( HOLE *h )
{
  const dReal *body_pos;
  dReal pos;
  dReal vel;
  dReal k_pos = 0.0;
  dReal k_vel = 1.0;
  dReal k_int = 0.0;
  dReal pos_desired = 0.0;
  dReal torque;

#ifdef COMMENT
  // commented out until the humanoid starts to play golf.
  // Freeze body
  dBodySetLinearVel( robot_bodies[ROOT], 0.0, 0.0, 0.0 );

  // Move club
  vel = dJointGetHingeAngleRate( club_joint );

  torque = - k_vel*vel;
  dJointAddHingeTorque( club_joint, torque );
#endif

  // no transition out
}

/****************************************************************/

int count = 0;

void control_robot( HOLE *h )
{
  int j;

  for ( j = 0; j < n_robot_joints; j++ )
    {
      joint_positions[j] = dJointGetHingeAngle( robot_joints[j] );
      joint_velocities[j] = dJointGetHingeAngleRate( robot_joints[j] );
      joint_torques[j] =
	- k_pos[j]*( joint_positions[j] - joint_positions_desired[j] )
	- k_vel[j]*joint_velocities[j];
      dJointAddHingeTorque( robot_joints[j], joint_torques[j] );
    }


  if ( count++ % 10000 == 0 )
    {
      j = j_ankle;
      printf( "%d %g %g %g\n", j, joint_positions[j], joint_velocities[j],
	      joint_torques[j] );

      j = j_knee;
      printf( "%d %g %g %g\n", j, joint_positions[j], joint_velocities[j],
	      joint_torques[j] );

      j = j_hip;
      printf( "%d %g %g %g\n", j, joint_positions[j], joint_velocities[j],
	      joint_torques[j] );
    }
}

/****************************************************************/
/****************************************************************/
/****************************************************************/
// Here is where we put in our player

void state_machine( HOLE *h )
{

  // return if we are stuck.
  if ( state == PLAN )
    {
      plan( h );
      move_club_back_waiting( h );
      control_robot( h );
    }
  else if ( state == MOVING_CLUB_BACK )
    {
      move_club_back( h );
      control_robot( h );
    }
  else if ( state == MOVING_INTO_SWING_POSITION )
   {
      move_into_swing_position( h );
      control_robot( h );
    }
  else if ( state == SWINGING )
    {
      swing_club( h );
      control_robot( h );
    }
  else if ( state == BALL_MOVING )
    {
      ball_moving( h );
      move_club_back_waiting( h );
      control_robot( h );
    }
  else if ( state == STUCK )
    {
      stuck( h );
      control_robot( h );
    }
  else
    {
      fprintf( stderr, "Unknown state %d in state_machine()\n", state );
      exit( -1 );
    }
}

/****************************************************************/
/****************************************************************/

void initialize( char *filename )
{
  int j;
  dReal erp, cfm;

  // setup pointers to drawstuff callback functions
  fn.version = DS_VERSION;
  fn.start = &start;
  fn.step = &simLoop;
  fn.stop = 0;
  fn.command = &command;
  fn.path_to_textures = DRAWSTUFF_TEXTURE_PATH;
 
  dInitODE ();
  // create world
  world = dWorldCreate( );
  space = dHashSpaceCreate( 0 );
  dWorldSetGravity( world, 0.0, 0.0, -9.81 );
  dWorldSetCFM( world, 1e-5 );
  // dWorldSetERP (dWorldID, dReal erp);
  erp = dWorldGetERP( world );
  cfm = dWorldGetCFM( world );
  /*
  printf( "erp: %g, cfm: %g, kp: %g, kd: %g\n",
	  erp, cfm, erp/(cfm*TIME_STEP), (1 - erp)/cfm );
  */

  contact_joint_group = dJointGroupCreate( 0 );
  robot_joint_group = dJointGroupCreate( 0 );

  create_bodies_and_geoms( filename, &current_hole );

  club_joint_ff = 0.0;

  for ( j = 0; j < MAX_N_ROBOT_PARTS; j++ )
    {
      joint_positions[j] = 0.0;
      joint_velocities[j] = 0.0;
      joint_torques[j] = 0.0;
      joint_positions_desired[j] = 0.0;
      k_pos[j] = 0.0;
      k_vel[j] = 0.0;
    }
  /*
  k_pos[ANKLE] = 100.0;
  k_vel[ANKLE] = 1.0;
  k_pos[KNEE] = 100.0;
  k_vel[KNEE] = 1.0;
  k_pos[HIP] = 100.0;
  k_vel[HIP] = 1.0;
  */
  k_pos[j_ankle] = 1000.0;
  k_vel[j_ankle] = 100.0;
  k_pos[j_knee] = 1000.0;
  k_vel[j_knee] = 100.0;
  k_pos[j_hip] = 1000.0;
  k_vel[j_hip] = 100.0;
  joint_positions_desired[j_ankle] = -0.08;
  joint_positions_desired[j_knee] = 0.2;
  joint_positions_desired[j_hip] = -0.2;
}

/****************************************************************/

void clean_up()
{

  dJointGroupDestroy( contact_joint_group );
  dJointGroupDestroy( robot_joint_group );
  dSpaceDestroy( space );
  dWorldDestroy( world );
  dCloseODE();
}

/****************************************************************/
/****************************************************************/

/*
Main program: When the program starts, the callbacks are set up,
everything is initialized, and then the simulation is started.
*/

int main( int argc, const char **argv )
{
  char *filename;

  if ( argc < 2 )
    filename = (char *) "c0000.txt";
  else
    filename = (char *) (argv[1]);
  printf( "Using hole file %s\n", filename );

  initialize( filename );

  // run simulation
  dsSimulationLoop( argc, (char **) argv, WINDOW_WIDTH, WINDOW_HEIGHT, &fn );

  clean_up();
  return 0;
}

/****************************************************************/
